/**
 * Copyright (C) 2022 Xilinx, Inc
 * Copyright (C) 2022 Advanced Micro Devices, Inc. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

// ------ I N C L U D E   F I L E S -------------------------------------------
// Local - Include Files
#include "ReportMemory.h"
#include "MAUtilities.h"

// 3rd Party Library - Include Files
#include <boost/lexical_cast.hpp>
#include <boost/property_tree/json_parser.hpp>

// ------ S T A T I C   V A R I A B L E S -------------------------------------
constexpr uint32_t no_sensor_dev        = 0xffffffff;
constexpr int      invalid_sensor_value = 0;

// ------ S T A T I C   F U N C T I O N S -------------------------------------
namespace {
template <typename T>
inline std::string pretty(const T& val, const std::string& default_val = "N/A", bool isHex = false) {
  if (typeid(val).name() != typeid(std::string).name()) {
    if (val >= std::numeric_limits<T>::max() || val == 0)
      return default_val;

    if (isHex) {
      std::stringstream ss;
      ss << "0x" << std::hex << val;
      return ss.str();
    }
  }
  return boost::lexical_cast<std::string>(val);
}

} //unnamed namespace

void ReportMemory::getPropertyTree(const std::shared_ptr<MA35Device> dev, boost::property_tree::ptree& _pt) const {
  boost::property_tree::ptree pt;

  XBUtilities::get_mem_info(dev, pt);

  // There can only be 1 root node
  _pt.add_child("memories", pt);
}

void ReportMemory::writeReport(const boost::property_tree::ptree& _pt, std::ostream& _output) const {
  boost::property_tree::ptree empty_ptree;

  _output << std::endl;
  if (_pt.get<std::string>("memories.status") != "ok") {
    _output << "ERROR " << _pt.get<std::string>("memories.status") << std::endl;
    return;
  }

  try {
    auto memories_pt = _pt.get_child_optional("memories.memories");
    if (memories_pt) {
      _output << "Memories:" << std::endl;
      _output << boost::format("    %-25s%-17s%-17s%-17s\n") % "     Tag" % "Size (MB)" % "Used (MB)" % "Used %";
      for (auto& v : memories_pt.get()) {
        const auto& t1     = v.second;
        auto        tag    = v.first;
        auto        size   = t1.get<std::string>("size_MB", "?");
        auto        used   = t1.get<std::string>("used_MB", "?");
        auto        used_p = t1.get<std::string>("used%", "?");
        _output << boost::format("  %-25s : %-17s%-17s%-17s\n") % tag % size % used % used_p;
      }
    }
    _output << "Memory Bandwidths:" << std::endl;
    //_output << boost::format("    %-25s%-20s%-20s\n") % "     Tag" % "Current (MBps)" % "Peak (MBps)";
    _output << boost::format("    %-25s%-20s\n") % "     Tag" % "Current (MBps)";
    for (auto& v : _pt.get_child("memories.bandwidth", empty_ptree)) {
      const auto& t1  = v.second;
      auto        tag = v.first;
      auto        cur = t1.get<std::string>("instantaneous", "?");
      //auto        peak = t1.get<std::string>("peak", "?");
      //_output << boost::format("  %-25s : %-20s%-20s\n") % tag % cur % peak;
      _output << boost::format("  %-25s : %-20s\n") % tag % cur;
    }
  } catch (std::exception const&) {
    _output << "ERROR: Unexpected error" << std::endl;
  }

  _output << std::endl;
}
