/**
 * Copyright (C) 2022 Xilinx, Inc
 * Copyright (C) 2022 Advanced Micro Devices, Inc. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

// ------ I N C L U D E   F I L E S -------------------------------------------
// Local - Include Files
#include "SubCmdCreateVF.h"
#include "../common/MA35Device.h"
#include "../common/MAUtilitiesCore.h"
#include "../common/MAUtilities.h"
namespace XBU = XBUtilities;

// 3rd Party Library - Include Files
#include <boost/program_options.hpp>
#include <boost/format.hpp>
namespace po = boost::program_options;

// System - Include Files
#include <iostream>
#include <set>

// ----- C L A S S   M E T H O D S -------------------------------------------

SubCmdCreateVF::SubCmdCreateVF(bool _isHidden, bool _isDepricated, bool _isPreliminary) : SubCmd("numvfs", "Create a VF. Or destroys the active VF") {
  const std::string longDescription = "Create a VF. Or destroys the active VF";
  setLongDescription(longDescription);
  setExampleSyntax("");
  setIsHidden(_isHidden);
  setIsDeprecated(_isDepricated);
  setIsPreliminary(_isPreliminary);
}

void SubCmdCreateVF::execute(const SubCmdOptions& _options) const {
  XBU::verbose("SubCommand: numvfs");
  // -- Retrieve and parse the subcommand options -----------------------------
  std::vector<std::string> devices_str;
  int32_t                  numvfs = -1;
  bool                     help   = false;

  po::options_description commonOptions("Common Options");
  commonOptions.add_options()("device,d", boost::program_options::value<decltype(devices_str)>(&devices_str)->multitoken(),
      "The Bus:Device.Function (e.g., 0000:d8:00.0) device of interest.")("value,v", boost::program_options::value<decltype(numvfs)>(&numvfs),
      "Number of VFs to create:\n"
      "  0         - Destroy active VFs\n"
      "  1         - Create one VF\n") ("help", boost::program_options::bool_switch(&help), "Help to use this sub-command");

  po::options_description hiddenOptions("Hidden Options");

  // Parse sub-command ...
  po::variables_map vm;
  process_arguments(vm, _options, commonOptions, hiddenOptions);

  // Check to see if help was requested or no command was found
  if (help) {
    printHelp(commonOptions, hiddenOptions);
    return;
  }

  // -- Now process the subcommand --------------------------------------------
  // Collect all of the devices of interest
  std::set<std::string> deviceNames;
  DeviceCollection      devices; // The collection of devices to examine
  for (const auto& deviceName : devices_str)
    deviceNames.insert(boost::algorithm::to_lower_copy(deviceName));

  try {
    XBU::collect_devices(deviceNames, true /*inUserDomain*/, devices);
  } catch (const std::runtime_error& e) {
    // Catch only the exceptions that we have generated earlier
    std::cerr << boost::format("ERROR: %s\n") % e.what();
    throw std::runtime_error("Operation Cancelled");
  }

  // enforce 1 device specification
  if (devices.empty() || devices.size() > 1) {
    std::cerr << "\nERROR: Please specify a single device using --device option\n\n";
    XBU::list_devices(std::cerr, true);
    throw std::runtime_error("Operation Cancelled");
  }

  // Get the device
  auto& working_device = devices[0];

  XBU::sudo_or_throw("Root privileges are required for numvfs cmd");

  if (!XBU::dev_present(working_device->bdf(), true)) {
    std::cerr << "****************************************************\n";
    std::cerr << "ERROR: Device not available" << std::endl;
    XBU::list_devices(std::cerr, true);
    std::cerr << "****************************************************\n";
    printHelp(commonOptions, hiddenOptions);
    throw std::runtime_error("Operation Cancelled");
  }
  if (numvfs < 0 || numvfs > 1) {
    std::cerr << "****************************************************\n";
    std::cerr << "ERROR: Num od vfs value must be either 0 or 1" << std::endl;
    std::cerr << "****************************************************\n";
    printHelp(commonOptions, hiddenOptions);
    throw std::runtime_error("Operation Cancelled");
  }

  // Ask user for permission
  if (!XBU::can_proceed(XBU::getForce()))
    throw std::runtime_error("Operation Cancelled");

  boost::property_tree::ptree ma35_pt1;
  XBU::do_numvfs(working_device, numvfs, ma35_pt1);
  if (ma35_pt1.get<std::string>("status") != "ok") {
    std::cerr << "****************************************************\n";
    if (numvfs == 1) {
      std::cerr << "ERROR: Create VF command failed" << std::endl;
    } else {
      std::cerr << "ERROR: Destroy VF command failed" << std::endl;
    }
    std::cerr << ma35_pt1.get<std::string>("status") << std::endl;
    std::cerr << "****************************************************\n";
  } else {
    std::cout << "****************************************************\n";
    if (numvfs == 1) {
      std::cout << "Create VF command completed" << std::endl;
    } else {
      std::cout << "Destroy VF command completed" << std::endl;
    }
    std::cout << "****************************************************\n";
  }
}
