// Copyright (C) 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#pragma once

#include <unistd.h>
#include <assert.h>
#include <sys/eventfd.h>

#include <filesystem>
#include <tuple>
#include <numeric> // for std:accumulate

#include <sys/ioctl.h>
#include <linux/vfio.h>
#include <thread>

#include "Device.h"

extern "C" {
#include <pci/pci.h>
}

#include <Logging.h>

static constexpr unsigned int BUFSIZE = 16384;
using namespace std::chrono_literals;

class VfioDevice : public Device {
  using clock_t = std::chrono::high_resolution_clock;

  enum class SubSys { VCE, VCDA, VCDB_1, VCDB_2, VCDB_3, GPU2D, VIP, XABR, XENC, XFPS, XAV1 };

  static constexpr uint32_t IATU_BASE_OFF = 0x0000'3100u;
  enum AtuReg { REGION_CTRL_1, REGION_CTRL_2, LWR_BASE_ADDR, UPPER_BASE_ADDR, LIMIT_ADDR, LWR_TARGET_ADDR, UPPER_TARGET_ADDR, REGION_CTRL_3 };
  static constexpr uint32_t REGION_ENABLE = 0x8000'0000u;

  static constexpr uint32_t TOP_SYS_CON_OFF             = 0x0080'0000;
  static constexpr uint32_t SC_MST_ADB_LPI_CON_STUS_OFF = TOP_SYS_CON_OFF + 0x508;
  static constexpr uint32_t SC_SLV_ADB_LPI_CON_STUS_OFF = TOP_SYS_CON_OFF + 0x50c;
  static constexpr uint32_t SYS_RST_CON_STUS_OFF        = TOP_SYS_CON_OFF + 0x100;

  static constexpr uint32_t PCIE_CONFIG_BASE = 0x0040'0000;

  static constexpr uint32_t FW_LOAD_ADDR    = 0x0290'0000;
  static constexpr uint32_t ZSP_SRAM_ADDR   = 0x0290'0000;
  static constexpr uint32_t CMD_DIR_ADDR    = 0x1fff4;
  static constexpr uint32_t CMD_RSN_ADDR    = 0x1fff8;
  static constexpr uint32_t SIGNATURE_TO_EP = 0x544F4550;
  static constexpr uint32_t SIGNATURE_TO_RC = 0x544F5243;
  static constexpr uint32_t GO_CMD          = 0x5a5a5a5a;
  static constexpr uint32_t DONE            = 0x444F4E45;
  static constexpr uint32_t SIG_ERROR       = 0x4641494C;

private:
  static constexpr struct {
    uint64_t eps;
    uint64_t epe;
    uint32_t off;
    uint32_t bar;
  } sm_barCfgs[] = {
      {0x4'0000'0000, 0x4'0090'0000, 0x0000'0000, 2},
      {0x4'0120'0000, 0x4'0130'0000, 0x0090'0000, 2},
      {0x4'1200'0000, 0x4'12A0'0000, 0x00B0'0000, 2},
      {0x4'1300'0000, 0x4'1350'0000, 0x0150'0000, 2},
      {0x4'1600'0000, 0x4'16A0'0000, 0x01A0'0000, 2},
      {0x4'1700'0000, 0x4'1750'0000, 0x0240'0000, 2},
      {0x4'1794'0000, 0x4'17D2'0000, 0x0290'0000, 2},
      // {0x4'0000'0000, 0x4'0090'0000, 0x02D0'0000, 2},
      {0x4'1900'0000, 0x4'1950'0000, 0x02D0'0000, 2},
      {0x4'1A00'0000, 0x4'1A50'0000, 0x0320'0000, 2},
      {0x0'0000'0000, 0x0'0800'0000, 0x0000'0000, 4},
  };
  static constexpr struct {
    uint32_t clockCtrl;
    uint32_t clockCtrlBits;
    uint32_t sysReset;
    uint32_t sysResetBits;
    uint32_t powerSetting;
    uint32_t powerSettingClearBits;
    uint32_t powerSettingPollingBits;
    uint32_t adbSetting;
    uint32_t adbSettingClearBits;
    uint32_t adbSettingPollingBits;
  } sm_subSysCfgs[] = {{0x14, 0x1, 0x114, 0x1, 0x710, 0x1, 0x2, 0x520, 0x3, 0xa0}, {0x00, 0x1, 0x100, 0x1, 0x700, 0x1, 0x2, 0x510, 0x1, 0x20},
      {0x04, 0x1, 0x104, 0x1, 0x704, 0x1, 0x2, 0x514, 0x1, 0x20}, {0x08, 0x1, 0x108, 0x1, 0x708, 0x1, 0x2, 0x518, 0x1, 0x20},
      {0x0c, 0x1, 0x10c, 0x1, 0x70c, 0x1, 0x2, 0x51c, 0x1, 0x20}, {0x1c, 0x1, 0x11c, 0x1, 0x720, 0x1, 0x2, 0x528, 0x1, 0x20},
      {0x18, 0x7, 0x118, 0x7, 0x724, 0x1, 0x2, 0x524, 0x3, 0xa0}, {0x24, 0x1, 0x124, 0x1, 0x71c, 0x1, 0x2, 0x52c, 0x1, 0x20},
      {0x20, 0x7, 0x120, 0x7, 0x718, 0x1, 0x2, 0x530, 0x3, 0xa00}, {0x28, 0x1, 0x128, 0x1, 0x714, 0x1, 0x2, 0x530, 0x3, 0xa00},
      {0x20, 0x1, 0x120, 0x1, 0x714, 0x1, 0x2, 0x60c, 0x1, 0x6}};

private:
  template <typename T = uint32_t>
  T Read(uint64_t Offset) {
    T    val;
    auto count = pread(m_device, &val, sizeof(val), Offset);
    if (count != sizeof(val)) {
      throw std::runtime_error("Read Failed!");
    }
    return val;
  }
  template <typename T = uint32_t>
  unsigned Write(uint64_t Offset, const T& Val) {
    auto count = pwrite(m_device, &Val, sizeof(Val), Offset);
    if (count != sizeof(Val)) {
      throw std::runtime_error("Write Failed!");
    }
    return count;
  }

  template <typename T = uint32_t>
  T ReadR(uint64_t Offset) {
    auto val = Read<T>(Offset);
    // logNotice("%08x:%08x %08x", uint32_t(Offset >> 32), uint32_t(Offset), val);
    return val;
  }

  template <typename T = uint32_t>
  void WriteRWR(uint64_t Offset, const T& Val) {
    auto init = Read<T>(Offset);
    Write(Offset, Val);
    auto read = Read<T>(Offset);
    if (read != init) {
      logNotice("%08x:%08x OLD %08x NEW %08x RB %08x%c", uint32_t(Offset >> 32), uint32_t(Offset), init, Val, read, " *"[read != init]);
    }
  }
  template <typename FUNC>
  bool WaitFor(const FUNC& Func, clock_t::duration Time = 1s, clock_t::duration Interval = 1ms) {
    auto end = clock_t::now() + Time;
    while (clock_t::now() <= end) {
      if (Func()) {
        return true;
      }
      std::this_thread::sleep_for(Interval);
    }
    return false;
  }

  auto GetBar(unsigned Index) {
    vfio_region_info region = {}; // shut up valgrind (doesn't know ioctls)
    region.argsz            = sizeof(region);
    region.index            = Index;
    if (ioctl(m_device, VFIO_DEVICE_GET_REGION_INFO, &region) < 0) {
      logFatal("VFIO_DEVICE_GET_REGION_INFO failed");
      fail();
    }
    return std::make_pair(region.offset, region.size);
  }
  void EnableBusMaster() {
    auto [configBase, configSize] = GetBar(VFIO_PCI_CONFIG_REGION_INDEX);
    uint16_t val                  = Read<uint16_t>(configBase + 4);
    Write<uint16_t>(configBase + 4, val | 4);
  }

  uint32_t CfgRegOff(unsigned Atu, AtuReg Reg) { return 0x3100 + Atu * 0x200 + Reg * 4; }
  void     Bar0Config(uint64_t Bar2Base, uint64_t Bar4Base);
  void     ResetZsp();
  // void  SetupInterruptFds();
  void PciConfig(); // Writes to DDR for ZSP, needed for firmware!
  void LoadFirmware();
  void ConfigureSubSys(SubSys SubSys, bool Enable);
  void EnableSubSys(SubSys SubSys) { ConfigureSubSys(SubSys, true); }
  // void DisableSubSys(SubSys SubSys) { ConfigureSubSys(SubSys, false); } // I don't think this works properly (in the driver as well)

public:
  explicit VfioDevice(const pci_dev& Device, bool Demo);
  void VfioBindPciDevice();
  void VfioConfigPciDevice();
  void Unbind_Bind(const std::string& Dpath, const std::string& Dubp);
  void GroupOwnership(const std::string& TargetDevSysfsIommuGrpPath);
  void VfioTestDevice();
};
