// Copyright (C) 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <pwd.h>

#include "VfioDevice.h"

void VfioDevice::VfioBindPciDevice() {
  // const auto target_dev_sysfs_path           = "/sys/bus/pci/devices/" + m_dbdf;
  const auto target_dev_sysfs_path           = "/sys/bus/pci/devices/" + GetBdf();
  const auto target_dev_sysfs_iommu_path     = target_dev_sysfs_path + "/iommu";
  const auto target_dev_sysfs_iommu_grp_path = target_dev_sysfs_path + "/iommu_group";
  const auto dev_sysfs_path                  = target_dev_sysfs_path + "/iommu_group/devices/";

  logInfo("target_dev_sysfs_path: %s", target_dev_sysfs_path.c_str());
  struct stat info;
  if (stat(target_dev_sysfs_path.c_str(), &info) != 0) {
    logFatal("Error: m_pciDevice %s does not exist, unable to bind device.", target_dev_sysfs_path.c_str());
    fail();
  }

  logDebug("target_dev_sysfs_iommu_path: %s", target_dev_sysfs_iommu_path.c_str());
  if (stat(target_dev_sysfs_iommu_path.c_str(), &info) != 0) {
    logFatal("Error: No signs of an IOMMU. Check your hardware and/or linux cmdline parameters. Use intel_iommu=on or iommu=pt iommu=1");
    fail();
  }

  logInfo("Binding...");
  // auto dsp   = dev_sysfs_path + m_dbdf;
  auto dsp   = dev_sysfs_path + GetBdf();
  auto dpath = dsp + "/driver";
  auto dop   = dsp + "/driver_override";
  auto dubp  = dpath + "/unbind";

  /* driver_override */
  auto fd = open(dop.c_str(), O_WRONLY);
  if (fd < 0) {
    logFatal("driver_override: file Not opened");
    fail();
  }
  auto ret = write(fd, "vfio-pci\n", sizeof("vfio-pci"));
  if (ret != sizeof("vfio-pci")) {
    logFatal("driver_override: write failed");
    fail();
  }
  close(fd);

  Unbind_Bind(dpath, dubp);

  // VFIO FLR Workaround (reset_method), avoid the FLR reset from VFIO.

  // auto        f_rst = "/sys/bus/pci/devices/" + m_dbdf + "/reset_method";
  auto        f_rst = "/sys/bus/pci/devices/" + GetBdf() + "/reset_method";
  std::string w_str;

  // if (w_str.empty()) {
  // logNotice("w_str EMPTY !!");
  // logNotice("rst_file: %s", f_rst.c_str());
  // logNotice("w_str size: %ld", w_str.size());
  // }

  auto fd_rst = open(f_rst.c_str(), O_WRONLY);
  if (fd_rst < 0) {
    logFatal("reset_method: file Not opened");
    fail();
  }
  auto ret_rst = write(fd_rst, w_str.c_str(), 1); // w_str.size();
  // logNotice("%ld", ret_rst);
  if ((unsigned int) ret_rst != 1) { // w_str.size()
    logFatal("reset_method: write failed");
    fail();
  }
  close(fd_rst);

  //

  GroupOwnership(target_dev_sysfs_iommu_grp_path);

  logInfo("success...");
  // logNotice("m_pciDevice 0x%x at %s bound to vfio-pci", GetPciDevice().vendor_id, GetBdf().c_str());
  logInfo("m_pciDevice 0x%x at %s bound to vfio-pci", GetPciDevice().vendor_id, GetBdf().c_str());
  logDebug("Devices listed in /sys/bus/pci/drivers/vfio-pci:");
  for (const auto& entry : std::filesystem::directory_iterator("/sys/bus/pci/drivers/vfio-pci")) {
    std::string filepath = entry.path();
    // if (filepath.std::string::find(m_dbdf) != std::string::npos) {
    if (filepath.std::string::find(GetBdf()) != std::string::npos) {
      logDebug(":\t%s", entry.path().filename().c_str());
    }
  }

  logDebug("ls -l /dev/vfio/");
  for ([[maybe_unused]] const auto& entry : std::filesystem::directory_iterator("/dev/vfio")) {
    logDebug(":\t%s", entry.path().filename().c_str());
  }
}

void VfioDevice::VfioConfigPciDevice() {
  auto container = open("/dev/vfio/vfio", O_RDWR);
  if (container < 0) {
    logFatal("Cannot create a new vfio container");
    fail();
  }
  if (ioctl(container, VFIO_GET_API_VERSION) != VFIO_API_VERSION) {
    logFatal("Unknown API version %d vs %d", VFIO_GET_API_VERSION, VFIO_API_VERSION);
    fail();
  }
  if (!ioctl(container, VFIO_CHECK_EXTENSION, VFIO_TYPE1_IOMMU)) {
    logFatal("IOMMU not supported");
    fail();
  }

  logInfo("iommu_group (VfioConfigPciDevice): %s", Device::m_iommu_grp.c_str());
  std::string iommu_grp_path = "/dev/vfio/" + Device::m_iommu_grp;
  auto        group          = open(iommu_grp_path.c_str(), O_RDWR);
  if (group < 0) {
    logFatal("Cannot open vfio group %s", iommu_grp_path.c_str());
    fail();
  }
  vfio_group_status groupStatus = {}; // shut up valgrind (doesn't know ioctls)
  groupStatus.argsz             = sizeof(groupStatus);
  if (ioctl(group, VFIO_GROUP_GET_STATUS, &groupStatus) < 0) {
    logFatal("VFIO_GROUP_GET_STATUS failed");
    fail();
  }
  if (!(groupStatus.flags & VFIO_GROUP_FLAGS_VIABLE)) {
    logFatal("Group is not viable");
    fail();
  }
  if (ioctl(group, VFIO_GROUP_SET_CONTAINER, &container) < 0) {
    logFatal("VFIO_GROUP_SET_CONTAINER failed");
    fail();
  }
  if (ioctl(container, VFIO_SET_IOMMU, VFIO_TYPE1_IOMMU) < 0) {
    logFatal("VFIO_SET_MMU failed: %d", errno);
    fail();
  }
  vfio_iommu_type1_info iommuInfo = {}; // shut up valgrind (doesn't know ioctls)
  iommuInfo.argsz                 = sizeof(iommuInfo);
  if (ioctl(container, VFIO_IOMMU_GET_INFO, &iommuInfo) < 0) {
    logFatal("VFIO_IOMMU_GET_INFO failed");
    fail();
  }
  logInfo("pgsizes: %llx", iommuInfo.iova_pgsizes);

  // m_device = ioctl(group, VFIO_GROUP_GET_DEVICE_FD, m_dbdf.c_str());
  m_device = ioctl(group, VFIO_GROUP_GET_DEVICE_FD, GetBdf().c_str());
  if (m_device < 0) {
    logFatal("VFIO_GROUP_GET_DEVICE_FD failed");
    fail();
  }

  // close(container);  // Is Required?
  // close(group);  // Is Required?

  // if (ioctl(m_device, VFIO_DEVICE_RESET) < 0) {	// Commented for workaround
  // logFatal("VFIO_DEVICE_RESET failed");
  // }

  EnableBusMaster();
  // Bar0Config(m_pciDevice.base_addr[2] & ~0xf, m_pciDevice.base_addr[4] & ~0xf);
  Bar0Config(GetPciDevice().base_addr[2] & ~0xf, GetPciDevice().base_addr[4] & ~0xf);
  std::tie(m_bar2Base, m_bar2Size) = GetBar(2);
  std::tie(m_bar4Base, m_bar4Size) = GetBar(4);

  // enable_interrupt_data_path
  // Write(m_bar2Base + 0x940, 0x9ffff1);
  // Write(m_bar2Base + 0x944, 0x0);
  // Write(m_bar2Base + 0x860220, 0x0);
  // Write(m_bar2Base + 0x860224, 0x0);

  // SetupInterruptFds(); // test only - non functional

  // PciConfig();
  // LoadFirmware(); // Uncommented: "PCIe Mode" ; Commented: "Normal Mode"
}

void VfioDevice::Bar0Config(uint64_t Bar2Base, uint64_t Bar4Base) {
  auto [bar0Base, bar0Size] = GetBar(0);
  auto count                = sizeof(sm_barCfgs) / sizeof(sm_barCfgs[0]);
  for (unsigned i = 0; i < count; ++i) {
    auto& config = sm_barCfgs[i];
    auto  eps    = config.eps;
    auto  epe    = config.epe;
    auto  base   = (config.bar == 2) ? Bar2Base : Bar4Base;
    auto  rcs    = base + config.off;
    auto  rce    = rcs + epe - eps - 1;
    Write(bar0Base + CfgRegOff(i, REGION_CTRL_2), 0x0);
    Write(bar0Base + CfgRegOff(i, REGION_CTRL_1), 0x0);
    Write(bar0Base + CfgRegOff(i, LWR_BASE_ADDR), uint32_t(rcs));           // RC start low address
    Write(bar0Base + CfgRegOff(i, UPPER_BASE_ADDR), uint32_t(rcs >> 32));   // RC start high address
    Write(bar0Base + CfgRegOff(i, LIMIT_ADDR), uint32_t(rce));              // RC end low address
    Write(bar0Base + CfgRegOff(i, LWR_TARGET_ADDR), uint32_t(eps));         // EP start low address
    Write(bar0Base + CfgRegOff(i, UPPER_TARGET_ADDR), uint32_t(eps >> 32)); // EP start high address
    Write(bar0Base + CfgRegOff(i, REGION_CTRL_2), REGION_ENABLE);
  }
}

void VfioDevice::ResetZsp() {
  auto addrMst = m_bar2Base + SC_MST_ADB_LPI_CON_STUS_OFF;
  Write(addrMst, Read(addrMst) & 0xfffffffe);
  if (!WaitFor([this, addrMst]() { return !(Read(addrMst) & (1 << 5)); })) {
    logFatal("Cannot reset ZSP (MST)");
    fail();
  }

  auto addrSlv = m_bar2Base + SC_SLV_ADB_LPI_CON_STUS_OFF;
  Write(addrSlv, Read(addrSlv) & 0xfffffffe);
  if (!WaitFor([this, addrSlv]() { return !(Read(addrSlv) & (1 << 5)); })) {
    logFatal("Cannot reset ZSP (SLV)");
    fail();
  }

  auto addrRst = m_bar2Base + SYS_RST_CON_STUS_OFF;
  Write(addrRst, Read(addrRst) & 0xffc04000);
  Write(addrRst, Read(addrRst) | 0x3fbfff);

  Write(addrMst, Read(addrMst) | 1);
  if (!WaitFor([this, addrMst]() { return Read(addrMst) & (1 << 5); })) {
    logFatal("Cannot reset ZSP (MST 2)");
    fail();
  }

  Write(addrSlv, Read(addrSlv) | 1);
  if (!WaitFor([this, addrSlv]() { return Read(addrSlv) & (1 << 5); })) {
    logFatal("Cannot reset ZSP (SLV 2)");
    fail();
  }
}

typedef union {
  u32 all;
  // struct {
  //   unsigned emulator_env       : 1;
  //   unsigned int_use_shadow_reg : 1;
  //   unsigned sub_sys_vcd        : 1;
  //   unsigned s1_vcd_a           : 1;
  //   unsigned s1_vcd_b1          : 1;
  //   unsigned s1_vcd_b2          : 1;
  //   unsigned s1_vcd_b3          : 1;
  //   unsigned s2_vcd_a           : 1;
  //   unsigned s2_vcd_b1          : 1;
  //   unsigned s2_vcd_b2          : 1;
  //   unsigned s2_vcd_b3          : 1;
  //   unsigned sub_sys_vce        : 1;
  //   unsigned s1_vce             : 1;
  //   unsigned s2_vce             : 1;
  //   unsigned vcmd_enable        : 1;
  //   unsigned sub_sys_xabr       : 1;
  //   unsigned sub_sys_xav1       : 1;
  //   unsigned sub_sys_gpu        : 1;
  //   unsigned s1_gpu             : 1;
  //   unsigned s2_gpu             : 1;
  //   unsigned sub_sys_vip        : 1;
  //   unsigned sub_sys_riscv      : 1;
  //   unsigned sys_riscv_0        : 1;
  //   unsigned sys_riscv_1        : 1;
  //   unsigned use_hdma_config    : 1;
  //   unsigned vip_cores          : 3;
  //   unsigned                    : 4;
  // } bit;
} pcie_config_t;

void VfioDevice::PciConfig() {
  pcie_config_t config;
  // config.all = 0;
  // config.bit.sub_sys_xav1 = 1;
  config.all                = 0x0205900a; // from driver
  auto [bar4Base, bar4Size] = GetBar(4);
  WriteRWR(bar4Base + PCIE_CONFIG_BASE, config.all);
}

/*
void VfioDevice::LoadFirmware() {
  const auto load = m_bar2Base + FW_LOAD_ADDR;
  const auto sram = m_bar2Base + ZSP_SRAM_ADDR;
  WriteRWR(load + CMD_DIR_ADDR, 0x33);
  // ResetZsp();
  if (!WaitFor([this, load]() { return !Read(load + CMD_DIR_ADDR); })) {
    logFatal("BootROM not complete");
    fail();
  }
  auto fd = open("/lib/firmware/supernova_zsp_fw.bin", O_RDONLY);
  if (fd < 0) {
    logFatal("Cannot open firmware");
    fail();
  }
  struct stat stat;
  fstat(fd, &stat);
  auto addr = static_cast<uint32_t*>(mmap(0, stat.st_size, PROT_READ, MAP_PRIVATE, fd, 0));
  if (!addr) {
    logFatal("Unabled to mmap firmware");
  }
  // weirdness on FPGA?
  auto off = load;
#if 1
  auto bytes = reinterpret_cast<uint8_t*>(addr);
  for (unsigned i = 0; i < stat.st_size; ++i, ++off) {
    WriteRWR(off, bytes[i]);
    std::this_thread::sleep_for(5us);
    if (!((i + 1) % 0x400)) {
      std::this_thread::sleep_for(100ms);
    }
  }
#else
  for (unsigned i = 0; i < stat.st_size / 4; ++i, off += 4) {
    WriteRWR(off, addr[i]);
  }
#endif
  // {
  //   uint32_t sum = 0;
  //   for (unsigned i = 0; i < stat.st_size; ++i) {
  //     sum += Read<uint8_t>(load + i);
  //   }
  //   logNotice("Firmware: %ld %08x", stat.st_size, sum);
  // }
  munmap(addr, stat.st_size);
  WriteRWR(load + CMD_DIR_ADDR, SIGNATURE_TO_EP);
  WriteRWR(sram + CMD_RSN_ADDR, GO_CMD);
  auto timeout = clock_t::now() + 10s;
  bool loaded  = false;
  while (clock_t::now() < timeout) {
    auto sig    = Read(sram + CMD_DIR_ADDR);
    auto status = Read(sram + CMD_RSN_ADDR);
    logNotice("SIG: %08x STATUS: %08x", sig, status);
    if (sig == SIGNATURE_TO_RC) {
      if (status == DONE) {
        loaded = true;
        break;
      } else if (status == SIG_ERROR) {
        logFatal("Signature check failed");
        fail();
      }
    }
    std::this_thread::sleep_for(1s);
  }
  if (!loaded) {
    logFatal("Firmware load timed out");
    // fail();
  }
}
*/

void VfioDevice::LoadFirmware() {
  unsigned long delay = 1000;
  const auto    load  = m_bar2Base + FW_LOAD_ADDR;
  const auto    sram  = m_bar2Base + ZSP_SRAM_ADDR;

  // ResetZsp();

  while (delay--) {
    if (Read(load + CMD_DIR_ADDR) == 0)
      break;
    std::this_thread::sleep_for(1020us);
  }
  if (Read(load + CMD_DIR_ADDR) != 0) {
    logFatal("bootrom pcie boot failed, please reboot host");
    fail();
  }

  // FILE* fd = fopen("/lib/firmware/supernova_zsp_fw_Telemetry.bin", "rb");
  FILE* fd = fopen("/lib/firmware/zsp_firmware_packed_pcie_Flash.bin", "rb");
  if (fd == NULL) {
    logFatal("Cannot open firmware file");
    fail();
  }
  fseek(fd, 0, SEEK_END);
  long unsigned int lSize = ftell(fd);
  rewind(fd);
  char* buffer = (char*) malloc(sizeof(char) * lSize);
  if (buffer == NULL) {
    logFatal("Memory error");
    fail();
  }

  // copy the file into the buffer
  size_t result = fread(buffer, 1, lSize, fd);
  if (result != lSize) {
    logFatal("Reading error");
    fail();
  }
  fclose(fd);

  auto      addr     = mmap(0, m_bar2Size, PROT_READ | PROT_WRITE, MAP_SHARED, m_device, m_bar2Base);
  uint32_t* ptr_load = reinterpret_cast<uint32_t*>((reinterpret_cast<uint8_t*>(addr)) + FW_LOAD_ADDR);

  memset(ptr_load, '0', result);
  memcpy(ptr_load, buffer, result); // Loading Firmware

  Write(sram + CMD_DIR_ADDR, SIGNATURE_TO_EP);
  Write(sram + CMD_RSN_ADDR, GO_CMD);
  std::this_thread::sleep_for(1s);

  auto timeout = clock_t::now() + 10s;
  bool loaded  = false;
  while (clock_t::now() < timeout) {
    auto sig    = Read(sram + CMD_DIR_ADDR);
    auto status = Read(sram + CMD_RSN_ADDR);
    logNotice("SIG: %08x STATUS: %08x", sig, status);
    if (sig == SIGNATURE_TO_RC) {
      if (status == DONE) {
        loaded = true;
        std::this_thread::sleep_for(1s);
        break;
      } else if (status == SIG_ERROR) {
        logFatal("Signature check failed");
        fail();
      }
    }
    std::this_thread::sleep_for(1s);
  }
  if (!loaded) {
    logFatal("Firmware load timed out");
    fail();
  }
}

void VfioDevice::ConfigureSubSys(SubSys SubSys, bool Enable) {
  if (SubSys == SubSys::VIP) {
    logFatal("VIP Unsupported");
    fail();
  }
  auto base = m_bar2Base + 0x1900000; // slice 0
  // auto base = m_bar2Base + 0x2800000; // slice 1
  auto cfg = sm_subSysCfgs[unsigned(SubSys)];
  if (Enable) {
    Write(base + cfg.clockCtrl, Read(base + cfg.clockCtrl) | cfg.clockCtrlBits);
    Write(base + cfg.sysReset, Read(base + cfg.sysReset) | cfg.sysResetBits);
    Write(base + cfg.powerSetting, Read(base + cfg.powerSetting) & ~cfg.powerSettingClearBits);
    if (!WaitFor([this, base, &cfg]() { return (Read(base + cfg.powerSetting) & cfg.powerSettingPollingBits) == cfg.powerSettingPollingBits; })) {
      logFatal("%u Power Configuration failed", unsigned(SubSys));
      fail();
    }
    if (SubSys == SubSys::XAV1) {
      Write(base + cfg.adbSetting, Read(base + cfg.adbSetting) & ~cfg.adbSettingClearBits);
      if (!WaitFor([this, base, &cfg]() { return Read(base + cfg.adbSetting) & cfg.adbSettingPollingBits; })) {
        logFatal("%u ADB Configuration failed", unsigned(SubSys));
        // fail();
      }
    } else {
      Write(base + cfg.adbSetting, Read(base + cfg.adbSetting) | cfg.adbSettingClearBits); // why are they called *clear* bits?
      if (!WaitFor([this, base, &cfg]() { return (Read(base + cfg.adbSetting) & cfg.adbSettingPollingBits) == cfg.adbSettingPollingBits; })) {
        logFatal("%u ADB Configuration failed", unsigned(SubSys));
        // fail(); // THIS IS HAPPENING IN THE DRIVER AS WELL!
      }
    }
    return;
  }
  Write(base + cfg.clockCtrl, Read(base + cfg.clockCtrl) & ~cfg.clockCtrlBits);
  Write(base + cfg.sysReset, Read(base + cfg.sysReset) & ~cfg.sysResetBits);
  Write(base + cfg.powerSetting, Read(base + cfg.powerSetting) | cfg.powerSettingClearBits);
  if (!WaitFor([this, base, &cfg]() { return !(Read(base + cfg.powerSetting) & 4); })) {
    logFatal("%u Power Configuration failed", unsigned(SubSys));
    fail();
  }
  if (SubSys == SubSys::XAV1) {
    Write(base + cfg.adbSetting, Read(base + cfg.adbSetting) | cfg.adbSettingClearBits);
    if (!WaitFor([this, base, &cfg]() { return Read(base + cfg.adbSetting) & cfg.adbSettingPollingBits; })) {
      logFatal("%u ADB Configuration failed", unsigned(SubSys));
      // fail();
    }
  } else {
    Write(base + cfg.adbSetting, Read(base + cfg.adbSetting) & ~cfg.adbSettingClearBits);
    if (!WaitFor([this, base, &cfg]() { return (Read(base + cfg.adbSetting) & cfg.adbSettingPollingBits) != cfg.adbSettingPollingBits; })) {
      logFatal("%u ADB Configuration failed", unsigned(SubSys));
      // fail(); // THIS IS HAPPENING IN THE DRIVER AS WELL!
    }
  }
}

void VfioDevice::Unbind_Bind(const std::string& Dpath, const std::string& Dubp) {
  struct stat info;
  if ((stat(Dpath.c_str(), &info) == 0) && (info.st_mode & S_IFDIR)) {
    char        curr_driver[PATH_MAX];
    std::string curr_driver_symlink;

    ssize_t len = ::readlink(Dpath.c_str(), curr_driver, sizeof(curr_driver) - 1);
    if (len == -1) {
      logFatal("readlink %s failed", Dpath.c_str());
      fail();
    }
    curr_driver[len]    = '\0';
    curr_driver_symlink = std::string(curr_driver);

    std::string curr_driver_name = std::filesystem::path(curr_driver_symlink).filename();
    logInfo("curr_driver: %s", curr_driver_name.c_str());
    if (curr_driver_name == "vfio-pci") {
      // logNotice("%s already bound to vfio-pci", GetBdf().c_str());
      logInfo("%s already bound to vfio-pci", GetBdf().c_str());
    } else {
      auto fd = open(Dubp.c_str(), O_WRONLY); // driver_unbind
      if (fd < 0) {
        logFatal("drivers_unbind: file Not opened");
        fail();
      }
      auto ret = write(fd, GetBdf().c_str(), GetBdf().size());
      if (ret != sizeof(GetBdf())) {
        logFatal("drivers_unbind: write failed");
        fail();
      }
      logNotice("Unbind(size): %ld", ret);
      close(fd);
      // logNotice("Unbound %s from %s", m_dbdf.c_str(), curr_driver);
      logNotice("Unbound %s from %s", GetBdf().c_str(), curr_driver);
    }
  } else {
    auto fd = open("/sys/bus/pci/drivers_probe", O_WRONLY);
    if (fd < 0) {
      logFatal("drivers_probe: file Not opened");
      fail();
    }
    auto ret = write(fd, GetBdf().c_str(), GetBdf().size());
    if ((unsigned int) ret != GetBdf().size()) {
      logFatal("drivers_probe: write failed");
      fail();
    }
    close(fd);
  }
}

void VfioDevice::GroupOwnership(const std::string& TargetDevSysfsIommuGrpPath) {
  char        iommu_group[PATH_MAX];
  std::string iommu_group_symlink;
  logInfo("TargetDevSysfsIommuGrpPath: %s", TargetDevSysfsIommuGrpPath.c_str());
  ssize_t len_grp = ::readlink(TargetDevSysfsIommuGrpPath.c_str(), iommu_group, sizeof(iommu_group) - 1);
  if (len_grp == -1) {
    logFatal("readlink %s failed", TargetDevSysfsIommuGrpPath.c_str());
    fail();
  }
  iommu_group[len_grp] = '\0';
  iommu_group_symlink  = std::string(iommu_group);

  logInfo("iommu_group_symlink: %s", iommu_group_symlink.c_str());
  std::string iommu_group_name = std::filesystem::path(iommu_group_symlink).filename();
  auto        iommu_group_path = "/dev/vfio/" + iommu_group_name;
  logInfo("iommu_group (VfioBindPciDevice): %s", iommu_group_name.c_str()); // To be passed to VfioDevice()
  Device::m_iommu_grp = iommu_group_name;

  auto   size = sysconf(_SC_GETPW_R_SIZE_MAX);
  size_t buf_size;
  buf_size = (size == -1) ? BUFSIZE : size_t(size);

  auto   buff = std::make_unique<unsigned char[]>(buf_size); // will throw on OOM
  passwd pwd;
  auto   result = &pwd;
  auto   s      = getpwnam_r(getenv("USER"), &pwd, reinterpret_cast<char*>(buff.get()), BUFSIZE, &result);
  if (!result) {
    if (!s) {
      logFatal("User Not found");
    } else {
      logFatal("getpwnam_r failed");
    }
    fail();
  }

  logInfo("User: %s ; UID: %d ; GID: %d", getenv("USER"), pwd.pw_uid, pwd.pw_gid);
  logInfo("iommu_group_path: %s", iommu_group_path.c_str());
  auto status = chown(iommu_group_path.c_str(), pwd.pw_uid, pwd.pw_gid);
  logInfo("chown_status=%d", status);
  if (status != 0) {
    logFatal("Error: unable to adjust group ownership of %s", iommu_group_path.c_str());
    fail();
  }
}

void VfioDevice::VfioTestDevice() {

  // Check Firmware
  auto val_dir = ReadR(m_bar2Base + ZSP_SRAM_ADDR + CMD_DIR_ADDR);
  auto val_st  = ReadR(m_bar2Base + ZSP_SRAM_ADDR + CMD_RSN_ADDR);
  if (val_dir != SIGNATURE_TO_RC || val_st != RUN_SUCCESS) {
    throw std::runtime_error("Firmware is Not ready, Initialization failed!");
  }

#if 0
  EnableSubSys(SubSys::XABR);
  uint32_t xabrVer = ReadR(m_bar2Base + 0xcc0000 + 0xb0);
  if (xabrVer != 0x8c8bfb9b) {
    logNotice("MA35D device initialization failed!");
    logNotice("XabrVer: 0x%08x", xabrVer);
    exit(1);
  }
  logNotice("XabrVer (VFIO): 0x%08x", xabrVer);

  // DisableSubSys(SubSys::XABR); // doesn't seem to work yet (investigate)
  // logNotice("XabrVer: 0x%08x", ReadR(m_bar2Base + 0xcc0000 + 0xb0));
  // EnableSubSys(SubSys::XENC);
  // EnableSubSys(SubSys::XFPS);
  // EnableSubSys(SubSys::XAV1);

  uint32_t spi_status2 = ReadR(m_bar2Base + 0x2a10000 + 0x1c);  // SPI-status2 (0x92)
  logNotice("Spi Status2: 0x%08x", spi_status2);

  logNotice("SPI0 - CTL3: 0x%08x", ReadR<uint8_t>(m_bar2Base + 0x2a10000 + 0x10));
  logNotice("SPI1: 0x%08x", ReadR(m_bar2Base + 0x2a11000 + 0x1c));
  logNotice("SPI2: 0x%08x", ReadR(m_bar2Base + 0x2a12000 + 0x1c));
#endif

#if 0
  uint32_t reset = 0x1000;
  pread(device, &reset, sizeof(reset), bar2_base + 0x2d00000 + 0x402094);
  do {
    pread(device, &fpsStat, sizeof(fpsStat), bar2_base + 0x2d00000 + 0x4020a8);
    logNotice("FPS_STATUS: %08x", fpsStat);
  } while (fpsStat & 0x3000000);

  logNotice("MBOX0_PTR: %08x", Read(m_bar2Base + 0x2d00000 + 0x401008));
  Write(m_bar2Base + 0x2d00000 + 0x401000, 23);
  logNotice("MBOX0_PTR: %08x", Read(m_bar2Base + 0x2d00000 + 0x401008));

  auto               addr = mmap(0, m_bar2Size, PROT_READ | PROT_WRITE, MAP_SHARED, m_device, m_bar2Base);
  volatile uint32_t* ptr  = reinterpret_cast<volatile uint32_t*>(reinterpret_cast<volatile uint8_t*>(addr) + 0x2d00000 + 0x401000);
  for (unsigned i = 0; i < 16; ++i, ++ptr) {
    logNotice("%p %08x", ptr, *ptr);
  }
  munmap(addr, m_bar2Size);
#endif
}

VfioDevice::VfioDevice(const pci_dev& PciDevice, bool Demo) : Device(PciDevice, Demo) {
  if (!IsDemo()) {
    VfioBindPciDevice();
    VfioConfigPciDevice();
    VfioTestDevice();
  }
}

template <typename T = uint32_t>
T Device::ReadBar2(uint64_t Offset) const {
  T    val;
  auto count = pread(m_device, &val, sizeof(val), (m_bar2Base + Offset));
  if (count != sizeof(val)) {
    throw std::runtime_error("ReadBar2 Failed!");
  }
  return val;
}
template uint32_t Device::ReadBar2<uint32_t>(uint64_t Offset) const;

template <typename T = uint32_t>
unsigned Device::WriteBar2(uint64_t Offset, const T& Val) {
  auto count = pwrite(m_device, &Val, sizeof(Val), (m_bar2Base + Offset));
  if (count != sizeof(Val)) {
    throw std::runtime_error("WriteBar2 Failed!");
  }
  return count;
}
template unsigned Device::WriteBar2<uint8_t>(uint64_t Offset, const uint8_t& Val);
template unsigned Device::WriteBar2<uint32_t>(uint64_t Offset, const uint32_t& Val);

template <typename T = uint32_t>
T Device::ReadBar4(uint64_t Offset) const {
  T    val;
  auto count = pread(m_device, &val, sizeof(val), (m_bar4Base + Offset));
  if (count != sizeof(val)) {
    throw std::runtime_error("ReadBar4 Failed!");
  }
  return val;
}
template uint32_t Device::ReadBar4<uint32_t>(uint64_t Offset) const;

template <typename T = uint32_t>
unsigned Device::WriteBar4(uint64_t Offset, const T& Val) {
  auto count = pwrite(m_device, &Val, sizeof(Val), (m_bar4Base + Offset));
  if (count != sizeof(Val)) {
    throw std::runtime_error("WriteBar4 Failed!");
  }
  return count;
}
template unsigned Device::WriteBar4<uint8_t>(uint64_t Offset, const uint8_t& Val);
template unsigned Device::WriteBar4<uint32_t>(uint64_t Offset, const uint32_t& Val);

std::unique_ptr<Device> Device::CreateDevice(const pci_dev& PciDevice, bool Demo) {
  return std::make_unique<VfioDevice>(PciDevice, Demo);
}
